import logging  
import os
from typing import Type
import json
import pandas as pd
import uuid
from pandas.core.frame import DataFrame
import pyodbc
from batch import plus_batch
from helper import plus_helper
from database import plus_database


#incoming Event (filename/url)

#class Batch (input= source_filename/url, 
#                  store: batchid, start_datetime, end_datetime, Source_filename, businessobjecttype, filetype, source_path, number_of_records, number_of_files, state
#                   generate: BatchId)
    #Generate BatchId
    #check if filename is correct and file exists
    #check if file was already handled
    #report (send message to queue BatchSummary)

#class SubBatch (input = batchid,
#                   store: subbatchid, start_datetime, end_datetime, target_filename, target_folder, 
#                           target_container, number_of_records, state
#                   generate: subbatchid, destination filename (input is filter in file and objecttype) and path,  )
    #Generate Subbatchid
    #Generate target filename and path
    #Report (Query now, API later)
    
class PlusSubbatch():
    def __init__(self):        
        self.destination_filetype=None
        self.number_of_records=0
        self.number_of_records_in_table = 0
        self.database=None
        self.database_schema=''
        self.database_table=''
        self.create_datetime=plus_helper.getcurrent_datetime()
        self.message_version=''

    def init_onramp(self, batch):
        if not isinstance(batch, plus_batch.PlusBatch):
            logging.info(f'Passed {type(batch)}, expecting object of Batch-type')
        self.cds_batchid=batch.cds_batchid
        self.source_filename=batch.source_filename
        self.source_folder=batch.source_folder
        self.business_object_type=batch.business_object_type

        self.destination_account=''
        self.destination_container=''
        self.destination_folder=''        

        self.batch=batch

        self.state = 'Init'       

        self.set_subbatchid()

    def set_database_connection(self, server, database_name, username, password):
        if isinstance(server, str) and isinstance(database_name, str) and isinstance(username, str) and isinstance(password, str):
            #Auto retry logic for when database is down is applied in Database module
            self.database=plus_database.PlusDatabase(server=server, name=database_name, username=username, password=password)
        else:
            self.database=None
            raise TypeError(f'One of the passed variables is not of type str.')

    def set_subbatchid(self, subbatch=''):
        if subbatch=='':
            self.cds_subbatchid = str(uuid.uuid4())
        else:
            self.cds_subbatchid = subbatch

    def set_target_filetype(self, dest_filetype='json'):
        if isinstance(dest_filetype, str):
            self.destination_filetype=dest_filetype
        else:
            raise TypeError(f'Expected str-input for dest_filetype variable.')

    def get_filename(self, split_attributes='', split_values=''):
        if split_attributes == '' or split_values == '':
            split_zip = ''
        elif isinstance(split_attributes, (str, tuple, list)) and isinstance(split_values, (str, tuple, list)):
            split_attr_new = [item.lower() for item in split_attributes]
            split_zip = zip(split_attr_new, split_values)
        else:
            logging.info('Expected empty values or values of type string, tuple or list for "split_attributes" and "split_values" variables.')
            return False
        
        split_zip_new=''
        for item in split_zip:
            split_zip_new+="_" + "_".join(item)

        #we expect the source filename to be formatted like this:
        # YYYY_{period_letter(P(eriode),W(eek),M(aand))}{PeriodeNumber(XX)}_{BusinessObject}.{filetype}
        #ie. 2021_W20_dvo-omzet.xlsx
        if not self.destination_filetype:
            self.set_target_filetype()

        self.year = self.source_filename.split('_')[0]
        self.period = self.source_filename.split('_')[1]
        self.destination_folder=f'{self.year}/{self.period}'
        self.destination_filename = f'{self.destination_folder}/{self.year}_{self.period}{split_zip_new}_{self.business_object_type}_{self.cds_subbatchid}.{self.destination_filetype}'

        return self.destination_filename
        
    def send_batchfiles_report(self, destination_type, action):
        #destination_type valid entries: msg_queue, direct_sql
        if destination_type == 'msg_queue':
            pass
        elif destination_type == 'direct_sql':
            if action == 'insert':
                sql = f"""INSERT INTO [dbo].[BatchFiles]
                        ([BatchId]
                        ,[BatchCreateDate]
                        ,[Filename]
                        ,[NumberOfRecords]
                        ,[State]
                        ,[NumberOfRecordsInTable]
                        ,[BatchEndDate]
                        ,[SubBatchId]
                        ,[Accountname]
                        ,[Containername]
                        ,[Foldername])
                        VALUES
                        (
                        '{self.batch.cds_batchid}'
                        ,'{self.create_datetime}'
                        ,'{self.destination_filename}'
                        ,{self.number_of_records}
                        ,'{self.state}'
                        ,0
                        ,NULL
                        ,'{self.cds_subbatchid}'
                        ,'{self.destination_account}'
                        ,'{self.destination_container}'
                        ,'{self.destination_folder}'
                        )"""
                self.database.execute_insert_query(sql)
            elif action == 'update':
                
                sql = f"""UPDATE [dbo].[BatchFiles]
SET [NumberOfRecords] = {self.number_of_records}
,[State] = '{self.state}'
,[NumberOfRecordsInTable] = {self.number_of_records_in_table}
,[BatchEndDate] = {f'{self.database.null_value}' if self.state not in ('Done', 'Error', 'Processed') else f"'{plus_helper.getcurrent_datetime()}'"}
WHERE SubBatchId = '{self.cds_subbatchid}'"""
                self.database.execute_insert_query(sql)

    def _get_onramp_msg_header(self):
        onramp_msg_header = {
                "messageType": self.batch.business_object_type,
                "creationDateTime": self.create_datetime,
                "Filename" : self.destination_filename,
                "Foldername" : self.destination_folder,
                "numberOfRecords" : self.number_of_records,
                "CDS_BatchId" : self.batch.cds_batchid, 
                "CDS_SubBatchId" : self.cds_subbatchid,
                "RapportageJaar" : self.year,
                "Rapportageperiode" : self.period[1:],
                "MessageVersion": self.message_version,
                "payload": []
            }
        return onramp_msg_header

    def _create_msg(self, df):
        self.msg = self._get_onramp_msg_header()

        records = json.loads(DataFrame(df).to_json(orient='records', date_format='iso'))
   
        for r in records:
            r['CDS_BatchId'] = self.batch.cds_batchid 
            r['CDS_SubBatchId'] = self.cds_subbatchid
            r['RapportageJaar'] = self.year
            r['Rapportageperiode'] = self.period[1:]
            record = {"record": r}
            
            payload = self.msg['payload']
            payload.append(record)
    
    def get_msg_with_data(self, df):
        if not isinstance(df, DataFrame):
            raise TypeError(f"Provided data in variable df is not a DataFrame. Can't process data.")
        
        self.number_of_records=df.shape[0]
        self.batch.number_of_records+= self.number_of_records
        self._create_msg(df=df)

        return self.msg
        
    def get_cds_notification_msg(self, action):
        self.notification_msg={
            "notification": {
                "MessageType": "cds-notificatie",
                "CDS_BatchId": self.cds_batchid,
                "CDS_SubBatchId": self.cds_subbatchid,
                "CreationDateTime": plus_helper.getcurrent_datetime(),
                "Accountname": self.destination_account,
                "Containername": self.destination_container,
                "Foldername": self.destination_folder,
                "Filename": self.destination_filename,
                "NumberOfRecords": self.number_of_records,
                "EtlRunStatus": self.state,
                "MessageVersion": self.message_version,
                "Action": action}
        }
        return json.dumps(self.notification_msg)

    def file_was_processed(self, blob):
        #a blob is expected to be a combination of folder and filename joined by a "/" ie. "2021/20/2021_P20_oos.xlsx"
        #the folder is optional.
        sql = f"SELECT COUNT(filename) as qty FROM BatchFiles WHERE filename = '{blob}' AND State IN ('Done')"
        result = self.database.execute_select_query(sql)

        if result:
            if result[0] > 0:
                return True
            else:
                return False
        else:
            logging.info('Something went wrong when checking if the file was processed, see previous error messages.')
    
    def get_batchfiles_info(self):
        if self.cds_subbatchid=='':
            raise ValueError('Cds_subbatchid not set, exiting process...')
        elif not self.database:
            raise ValueError('Database connection not set, use Class Method "set_database_connection()"...')
        else:
            sql = f"""SELECT [Id]
      ,[BatchId]
      ,[BatchCreateDate]
      ,[NumberOfRecords]
      ,[State]
      ,[NumberOfRecordsInTable]
      ,[BatchEndDate]
      ,[Accountname]
      ,[Containername]
      ,[Foldername]
      ,[Filename]
  FROM [dbo].[BatchFiles]
  WHERE SubBatchId = '{self.cds_subbatchid}' 
  AND state = 'ReadyForPickup'"""
            result = self.database.execute_select_query(sql)

            if not result:
                logging.info(f"No result returned from select query, can't continue...")
            else:
                logging.info('Setting some variables in de subbatch...')
                self.cds_batchid=result[1]
                self.number_of_records = result[3]
                self.destination_account = result[7]
                self.destination_container = result[8]
                self.destination_folder = result[9]
                self.destination_filename = result[10]
            


from typing import Type
import ast
import pandas as pd
import logging
from helper import plus_helper
import datetime
from pandas.core.base import DataError

class PlusDataFrame():
    def __init__(self, df, raw=False):
        self.df = df
        if raw:
            self.is_raw = True
            self.headers_cleaned = False
            self.headers_renamed = False
            self.default_values_applied = False
            self.unwanted_columns_removed = False
            self.empty_records_removed = False
            self.datatyped = False

            self.unwanted_columns=[]
            self.empty_records_columns=[]
            self.header_rename_dict={} 
            self.default_values_dict={}
            self.datatypes_dict={}
        else:
            self.is_raw = False                   
    
    @plus_helper.measure_runtime
    def clean_header_names(self):        
        self.df.columns = self.df.columns.str.replace('%', '')            
        self.df.columns = self.df.columns.str.replace(' ', '')            
        self.df.columns = self.df.columns.str.replace('/', '')            
        self.df.columns = self.df.columns.str.replace('\n', '')            
        self.df.columns = self.df.columns.str.replace('\u00eb', 'e')            
        self.df.columns = self.df.columns.str.replace('\u00e9', 'e')            
        self.df.columns = self.df.columns.str.replace('-', '')            
        self.df.columns = self.df.columns.str.replace('.', '')            
        self.df.columns = self.df.columns.str.replace('_', '')
 

    @plus_helper.measure_runtime
    def remove_unwanted_columns(self):
        if not isinstance(self.unwanted_columns, (list, str)):
            logging.info('Variable "unwanted_columns" is expected to be of type list or string')
            raise TypeError(f'"columns" variable is of type {type(self.unwanted_columns)}, should be of type list or string.')

        #expecting a list or string in the environment variable
        #columns = os.environ['plus_bi_columns_to_be_removed'].split(';')        
        for col in self.unwanted_columns:
            try:
                self.df.drop(col, axis=1, inplace=True)
            except Exception as e:
                logging.info(f'Exception raised: {e}')    

    @plus_helper.measure_runtime
    def remove_empty_records(self):
        #Removal of values should be performed on preferably the Primary Key of the dataset, this can be 1 or many columns.
        
        if not isinstance(self.empty_records_columns, (list, str)):
            logging.info('Variable "columns" is expected to be of type list or string')
            raise TypeError(f'"columns" variable is of type {type(self.empty_records_columns)}, should be of type list or string.')
        
        #expecting a list or string in the environment variable
        #columns = ast.literal_eval(os.environ["plus_bi_excel_column_for_empty_record_removal"])
        #The value(s) should be the original headername(s).

        try:
            if isinstance(self.empty_records_columns, str):
                self.df.dropna(subset=[self.empty_records_columns], inplace=True)
            else:
                self.df.dropna(subset=self.empty_records_columns, inplace=True)
        except Exception as e:
            logging.info(f'Column with name {self.empty_records_columns} not found; not filtering on empty records.\nError message: {e}')

    @plus_helper.measure_runtime
    def rename_headers(self):
        #Expected dict-format: {Old-name : New-name} ie. {wnknr : Winkelnummer}
        if not isinstance(self.header_rename_dict, dict):     
            raise TypeError(f'Type of variable "header_dict" should be of type dict. Provided type: {type(self.header_rename_dict)}.')
        
        #expecting a dict in the environment variable
        #header_dict = ast.literal_eval(os.environ["plus_bi_excel_new_header_names"])

        #first we check if all columns are present.
        for expected_column, v in self.header_rename_dict.items():
            try:
                if not expected_column in self.df.columns:                
                    raise AssertionError(f"Did not find column: {expected_column}")
            except Exception as e:
                    logging.error(f'Exception raised: {e}')

        logging.info('All columns present, renaming them.')
        self.df.rename(columns=self.header_rename_dict, inplace=True)

    def add_default_values(self):
        #Expected dict-format: {New-column-name : default_value} ie. {Winkelnummer : 0}
        #You can also add columns that do not exist yet, for example MessageVersion for non-POS dataflows.
        if not isinstance(self.default_values_dict, dict):     
            raise TypeError(f'Type of variable "default_values_dict" should be of type dict. Provided type: {type(self.default_values_dict)}.')

        #expecting a dict in the environment variable with the following format:
        #var_dict = var_dict = {'Date' : 'datetime.date.today()', 'DateTime' : 'datetime.datetime.now()' , 'Integer' : '0', 'Float' : '0.0', 'String' : 'Something'}
        #!!NOTE: for date(time) use the functions used in the example.
        try:            
            for attr, default_value in self.default_values_dict.items():
                logging.info(f'Adding default value for attribute "{attr}".')
                try:
                    #first we check if the string can be evaluated as another Class 
                    try:
                        self.df[attr].fillna(ast.literal_eval(default_value), inplace=True)
                    except KeyError:
                        self.df[attr] = ast.literal_eval(default_value)
                except Exception as e:
                    try:
                        #next we check if the default is a function
                        try:
                            self.df[attr].fillna(eval(default_value), inplace=True)
                        except KeyError:
                            self.df[attr] = eval(default_value)
                    except Exception as e:
                        #last we type it from the value itself (mainly str)
                        try:
                            self.df[attr].fillna(default_value, inplace=True)
                        except KeyError:
                            self.df[attr] = default_value
        except Exception as e:
            logging.error(f'Unable to apply default value for attribute: {e}') 

    def set_datatypes(self):
        #Expected dict-format: {New-column-name : datatype} ie. {Winkelnummer : 'int'}
        #NOTE: ALL columns should be included in this dict.
        if not isinstance(self.datatypes_dict, dict):     
            raise TypeError(f'Type of variable "datatypes_dict" should be of type dict. Provided type: {type(self.datatypes_dict)}.')

        try:
            #expecting a dict in the environment variable with the following format:
            #var_dict = var_dict = {'CreatieDatum' : 'str' , 'BonDatumTijd' : 'str', 'Bedrag' : 'float', 'Aantal' : 'int', 'Omschrijving' : 'str'}            
        
            if not all(key in self.df.columns for key in self.datatypes_dict.keys()):
                for key in self.datatypes_dict.keys():
                    if key not in self.df.columns:
                        raise KeyError(f'Key {key} in datatype dict not present in Dataframe Columns.')        

            self.df = self.df.astype(self.datatypes_dict)
        except Exception as e:
            logging.error(f'Something went wrong datatyping the dataframe.\nError: {e}') 
            raise RuntimeError("Couldn't datatype the dataframe, see previous errors...")

    @plus_helper.measure_runtime
    def clean_raw_dataframe(self):
        try:
            if self.is_raw:
                if not self.headers_cleaned:
                    self.clean_header_names()
                    self.headers_cleaned = True
                
                if not self.unwanted_columns_removed:
                    if len(self.unwanted_columns) > 0:
                        self.remove_unwanted_columns()
                    self.unwanted_columns_removed = True                

                if not self.empty_records_removed:
                    if len(self.empty_records_columns) > 0:
                        self.remove_empty_records()
                    self.empty_records_removed = True
                
                if not self.headers_renamed:
                    if len(self.header_rename_dict) > 0:
                        self.rename_headers()
                    self.headers_renamed = True

                if not self.default_values_applied:
                    if len(self.default_values_dict) > 0:
                        self.add_default_values()
                    self.default_values_applied = True
                
                if not self.datatyped:
                    if len(self.datatypes_dict) > 0:
                        self.set_datatypes()
                        self.datatyped = True
                    else:
                        raise DataError(f'No key-value pairs in Datatype dict detected.')
                logging.info('All cleaning operations successful, dataframe is clean.')
                self.is_raw = False
            else:
                logging.info('Dataframe is already cleaned.')

        except Exception as ex:
            logging.error(f'Something went wrong while cleaning the dataframe.')
            logging.error(f'Error message: {ex}')
            raise RuntimeError(f'Error during cleaning of dataframe, see previous errors...')

        
        





import logging  
import os
import pyodbc
import uuid
import json
from helper import plus_helper
from database import plus_database

#incoming Event (filename/url)

#class Batch (input= source_filename/url, 
#                  store: batchid, start_datetime, end_datetime, Source_filename, businessobjecttype, filetype, source_path, number_of_records, number_of_files, state
#                   generate: BatchId)
    #Generate BatchId
    #check if filename is correct and file exists
    #check if file was already handled
    #report (send message to queue BatchSummary)

#class SubBatch (input = batchid,
#                   store: subbatchid, start_datetime, end_datetime, target_filename, target_folder, 
#                           target_container, number_of_records, state
#                   generate: subbatchid, destination filename (input is filter in file and objecttype) and path,  )
    #Generate Subbatchid
    #Generate target filename and path
    #Report (Query now, API later)
    
class PlusBatch():
    def __init__(self, cds_batchid=''):
        self.cds_batchid = cds_batchid 
        self.number_of_records = 0         
        self.state = ''
        
        self.batch_start_datetime = plus_helper.getcurrent_datetime()
        self.report=None

    def init_onramp(self, file_url, business_object_type):
        if not self._is_valid_url(file_url):
            raise ValueError(f'Provided uri is invalid: {file_url}')
        if not self._has_valid_ext():
            raise AssertionError(f'Provided filetype is not supported: {self.filetype}')
        self.cds_batchid = str(uuid.uuid4())
        self.number_of_files = 0
        self.state = 'Init'
        self.business_object_type=business_object_type

        self.service_name=f'plus-bi-{business_object_type}-onramp-storageblob'
        self.message_type=f'plus-bi-{business_object_type}'

        self._set_storage_info()

    def _is_valid_url(self, file_url):
        if 'https://' in file_url.lower():
            self.url=file_url
            return True
        else:
            return False

    def _has_valid_ext(self):
        extensions = ('xlsx','xls','json','csv','parquet','xml')
        self.filetype = self.url.split('.')[-1]
        if self.filetype in extensions:
            return True
        else:
            return False
    
    def _has_correct_filename(self):
        # check if filename contains a year, a period and the expected businesstype
        logging.info(f'Checking if filename ({self.source_filename}) is valid.')
                
        filenamearray = self.source_filename.split('_')

        if len(filenamearray) < 3:
            return False

        self.year = filenamearray[0]
        try: 
            year = int(filenamearray[0])
            if year < 2099 and year >= 2000:
                logging.info('Valid year provided in filename.')
        except Exception as e:
            logging.error(f'{e}: {year} is not a valid year in {self.source_filename}')
        
        self.period = filenamearray[1]
        try: 
            if not self.period.startswith(('P','W','M')):
                raise AssertionError(f'Unexpected value where period is expected in filename. Value: {self.period}, should start with "P"(Period),"W"(week) or"M"(Month).')
        except Exception as e:
            logging.error(f'{e}: {self.period} is not a valid period in {self.source_filename}')
        
        businesstype = str(filenamearray[2]).lower()
        if self.business_object_type == '':
            raise AssertionError(f'Business Object Type not set, first use method "set_business_object_type(self, business_object_type)" in this Class.')
        if not businesstype.startswith(self.business_object_type):
            logging.error(f'{businesstype} is not of the expected value: {self.business_object_type}')
            return False
        
        file_extension_array = self.source_filename.split('.')
        if isinstance(file_extension_array, (list, tuple)):
            self.file_extension=str(file_extension_array[-1]).lower()
        elif isinstance(file_extension_array, str):
            logging.info(f'It looks like there is no file extension present: {self.source_filename}')
            return False
        else:
            logging.info(f'Unexpected type for file_extension_array: {type(file_extension_array)}')
            return False

        return True

    def _set_storage_info(self):
        self.source_filename=self.url.split('/')[-1]

        if not self._has_correct_filename():
            raise AssertionError(f'Filename is not as expected.\nFilename should be in format "YYYY_PXX_BusinessObjectType.FileExt".')

        self.source_url_account=plus_helper.split_custom_char(self.url, '/', 3)
        self.source_account=self.source_url_account.split('.')[0].split('/')[-1]
        self.source_container_and_folder=self.url.replace(self.source_url_account, '').replace(self.source_filename, '')
        self.source_container=self.source_container_and_folder.split('/')[1]
        self.source_folder='/'.join(self.source_container_and_folder.split('/')[2:])[:-1]

        self.destination_folder=self.year + '/' + self.period

    def _set_onramp_filename(self, dest_filename):
        #during processing we split the sourcefile based on one or more attributes.
        #On subbatch level we determine the filename
        pass

    def set_database_connection(self, server, database_name, username, password):
        if isinstance(server, str) and isinstance(database_name, str) and isinstance(username, str) and isinstance(password, str):
            #Auto retry logic for when database is down is applied in Database module
            self.database=plus_database.PlusDatabase(server=server, name=database_name, username=username, password=password)
        else:
            self.database=None
            raise TypeError(f'One of the passed variables is not of type str.')

    def file_was_processed(self):
        logging.info('Check if file %s was already processed.', self.source_filename)
        sql = f"SELECT COUNT(filename) as qty FROM BatchSummary WHERE filename like '%{self.source_filename}'"
        
        result = self.database.execute_select_query(sql)

        if result:
            if result[0] > 0:
                return True
            else:
                return False
        else:
            logging.info('Something went wrong when checking if the file was processed, see previous error messages.')

    def set_report_message(self, action):
        self.report = {
                    "Batch": { 
                      "BatchId": self.cds_batchid,
                      "Action": action,
                      "BatchCreateDate": self.batch_start_datetime ,
                      "Filename": self.source_filename, #has to be without path
                      "NumberOfRecords": self.number_of_records,
                      "NumberOfFiles": self.number_of_files,
                      "State": self.state,
                      "Accountname": self.source_account,
                      "Containername": self.source_container,
                      "Foldername": self.source_folder
                      }
                  }
        return self.report

    def send_report(self, destination_type):
        #destination_type valid entries: msg_queue, direct_sql

        if not self.report:
            self.set_report_message(action='insert')
        
        batch_report = self.report['Batch']
        if destination_type == 'msg_queue':
            pass
        elif destination_type == 'direct_sql':
            if batch_report['Action'].lower() == 'insert':
                sql = f"""INSERT INTO [dbo].[BatchSummary]
                        ([BatchId]
                        ,[BatchCreateDate]
                        ,[Filename]
                        ,[NumberOfRecords]
                        ,[NumberOfFiles]
                        ,[State]
                        ,[Accountname]
                        ,[Containername]
                        ,[Foldername])
                        VALUES
                        ('{self.cds_batchid}'
                        ,'{batch_report['BatchCreateDate']}'
                        ,'{batch_report['Filename']}'
                        ,{batch_report['NumberOfRecords']}
                        ,{batch_report['NumberOfFiles']}
                        ,'{batch_report['State']}'
                        ,'{batch_report['Accountname']}'
                        ,'{batch_report['Containername']}'
                        ,'{batch_report['Foldername']}')"""
                self.database.execute_insert_query(sql)                
            elif batch_report['Action'].lower() == 'update':
                sql = f"""UPDATE [dbo].[BatchSummary]
                        SET [NumberOfRecords] = {batch_report['NumberOfRecords']}
                        ,[NumberOfFiles] = {batch_report['NumberOfFiles']}
                        ,[State] = '{batch_report['State']}'
                        WHERE [BatchId] = '{self.cds_batchid}'"""
                self.database.execute_update_query(sql)

    def create_error_message(self, error_code, error_description):
        self.error_msg = { "BatchId": self.cds_batchid,
                            "MessageId": str(uuid.uuid4()),
                            "MessageType" : self.message_type + '_Error',
                            "ErrorDateTime" : plus_helper.getcurrent_datetime(),
                            "BatchFilename": self.source_filename,
                            "ErrorFilename":"",
                            "Errors":[{
                                "Servicename": self.service_name,
                                "ErrorCode": error_code,
                                "ErrorDescription": error_description
                            }],
                            "payload" : []
                            }
        return self.error_msg



import logging
import time 
from datetime import datetime, timedelta
import json
import base64

def auto_retry(max_tries=5, wait_time_in_seconds=10):
    def check(func):    
        def wrapper(*args, **kwargs):
            tries = 1
            retry = True
            while retry:      
                result_of_function = func(*args, **kwargs)
                print(f"{func.__name__}({args}, {kwargs}) Try #{tries} returned: {result_of_function}")            
                if result_of_function:
                    return result_of_function
                elif tries == max_tries:
                    return None
                else:
                    logging.info(f"{func.__name__}() returned: {result_of_function}, sleeping for {wait_time_in_seconds} seconds for next try.")
                    tries += 1
                    time.sleep(wait_time_in_seconds)            
        return wrapper
    return check

def measure_runtime(func):
    #decorator function, returns the total runtime of a decorated function or method.
    def wrapper(*args, **kwargs):
        starttime = time.time()
        result_of_function = func(*args, **kwargs)
        logging.info(f"{func.__name__}() function took " + str(time.time()-starttime) + " seconds to run")
        return result_of_function
    return wrapper

def split_custom_char(string, split_char, occurence):
# rnd_str = 'Https://doe-maar-iets.net/een-container-of-iets/folder-dan/file-naam.xlsx'
# tst = split_custom_char(rnd_str,'/', 3)
# print(tst)
# >> Https://doe-maar-iets.net

    if string.count(split_char) == 1:
        return string.split(split_char)[0]
    return split_char.join(string.split(split_char, occurence)[:occurence])

def getcurrent_datetime():
      now = datetime.now()
      return now.strftime("%Y/%m/%d %H:%M:%S")

def encode_data(d):
    json_data = json.loads(d)
    logging.debug('Type: %s', type(json_data))
    s = json.dumps(json_data)
    enc = s.encode()  # utf-8 by default
    payload = base64.encodestring(enc)
    logging.debug('encoded: %s', payload)
    return payload



import logging  
import os
from azure.storage.blob import generate_blob_sas, BlobServiceClient, BlobClient, ContainerClient, AccountSasPermissions
from datetime import datetime, timedelta
import json

class PlusStorage():
    def __init__(self, storage_cnxn_string):
        #incoming: 'DefaultEndpointsProtocol=https;AccountName=plsdevcdsdata;AccountKey=JsvdRXFZvhprZx/34/dwiP0gsN44hUyvz9mhK0yHelu7cMFRpwtvyyxepXY9wW97O5CzLgnD4yXiIjCsM61I/Q==;EndpointSuffix=core.windows.net'
        string_array = storage_cnxn_string.split(';')
        if len(string_array) == 1:
            logging.info(f'Invalid connection string provided for storage account:\n{storage_cnxn_string}')
        else:
            self.default_endpoints_protocol=string_array[0].replace('DefaultEndpointsProtocol=','')
            self.account_name=string_array[1].replace('AccountName=','')
            self.account_key=string_array[2].replace('AccountKey=','')
            self.endpoint_suffix=string_array[3].replace('EndpointSuffix=','')
            try:
                self.blob_service_client=BlobServiceClient.from_connection_string(storage_cnxn_string)
            except Exception as e:
                logging.info(f"Can't establish connection through provided connection string:\n{storage_cnxn_string}")
        
    def get_url_for_blob(self, container='', folder='', filename=''):
        if isinstance(container, str) and isinstance(folder, str) and isinstance(filename, str):       
            if container == '' or filename == '':
                logging.info(f'Need to provide at least a container and filename.')
                return None
            if folder != '':
                if folder.endswith('/'):
                    blob_name = folder + filename
                else:
                    blob_name = folder + '/' + filename
            else:
                blob_name = filename
        try:
            sas_token_blob = generate_blob_sas(
                                                account_name=self.account_name,
                                                account_key=self.account_key,
                                                container_name= container,
                                                blob_name = blob_name,
                                                snapshot=None,
                                                permission=AccountSasPermissions(read=True),
                                                expiry=datetime.utcnow() + timedelta(hours=1)
                                                )
            url = 'https://' + self.account_name + '.blob.core.windows.net/' + container + '/' + blob_name + '?' + sas_token_blob
            return url
        except Exception as ex:
            logging.error('Error 0: %s', ex.args[0])
            logging.error('Error 1: %s', ex.args[1])
            return None
    
    def upload_file(self, data, container, blob): 
        #blob is expected to be of form: {optional: folder}/{mandatory:filename with subbatchid} ie. 2021/06/2021_P06_oos_{subbatchid}.json               
        try:
            blob_client = self.blob_service_client.get_blob_client(container=container, blob=blob)
            blob_client.upload_blob(data, overwrite=True)
        except Exception as e:
            logging.error('Trying to upload error %s', e)
            logging.error(e.args[0])

    def get_data_as_text(self, datastream, encoding='Utf-8'):
        #blob is the combination of folder and filename
        #this will return the data inside the blob as text (mainly used when loading csv)
        return datastream.content_as_text(encoding=encoding)

    def get_data_as_bytes(self, datastream):
        #blob is the combination of folder and filename
        #this will return the data inside the blob as bytes
        return datastream.content_as_bytes()

    def download_file(self, container, blob):
        try:
            blob_client =BlobClient.from_blob_url(self.get_url_for_blob(container=container, folder="/".join(blob.split("/")[0:-1]), filename=blob.split("/")[-1]))
            return blob_client.download_blob()
        except Exception as ex:
            logging.error(f'Error: {ex}\nFailed to download Blob: {blob}\nContainer: {container}\nAccount: {self.account_name}')






